@@ -2,13 +2,15 @@
  *
  *	A Tcl wrapper for libxml2.
  *
+ * Copyright (c) 2005 explain
+ * http://www.explain.com.au
  * Copyright (c) 2003-2004 Zveno Pty Ltd
  * http://www.zveno.com/
  *
  * See the file "LICENSE" for information on usage and
  * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * $Id: tcllibxml2.c,v 1.6 2004/07/09 03:09:32 balls Exp $
+ * $Id: tcllibxml2.c,v 1.10 2005/11/04 06:23:04 balls Exp $
  */
 
 #include <tclxml-libxml2/tclxml-libxml2.h>
@@ -18,6 +20,7 @@
 #include <libxml/xmlreader.h>
 #include <libxml/SAX2.h>
 #include <libxml/xmlerror.h>
+#include <libxml/uri.h>
 #include <string.h>
 
 #define TCL_DOES_STUBS \
@@ -40,8 +43,6 @@ typedef struct TclXMLlibxml2Info {
 
   TclXML_Info *xmlinfo;		/* Generic data structure */
 
-  xmlExternalEntityLoader defaultLoader; /* default loader overridden while parsing */
-
   Tcl_HashTable *scope;		/* XML namespaces in scope */
 
 } TclXMLlibxml2Info;
@@ -52,10 +53,6 @@ typedef struct TclXMLlibxml2Info {
 
 static ClientData ReaderCreate _ANSI_ARGS_((Tcl_Interp *interp, 
 											TclXML_Info *xmlinfo));
-/*
-static ClientData TclXMLlibxml2CreateEntityParser _ANSI_ARGS_((
-                    Tcl_Interp *interp, ClientData clientData));
-*/
 static int  ReaderReset _ANSI_ARGS_((ClientData clientData));
 static int	TclXMLlibxml2Delete _ANSI_ARGS_((ClientData clientData));
 static int	ReaderParse _ANSI_ARGS_((ClientData clientData, 
@@ -66,18 +63,24 @@ static int	TclXMLlibxml2Configure _ANSI_ARGS_((ClientData clientdata,
 static int	TclXMLlibxml2Get _ANSI_ARGS_((ClientData clientData, 
                     int objc, Tcl_Obj *CONST objv[]));
 
+static xmlParserInputPtr TclXMLlibxml2ExternalEntityLoader _ANSI_ARGS_((const char *URL,
+									const char *ID,
+									xmlParserCtxtPtr ctxt));
+
 /*
  * Externally visible functions
  */
 
 typedef struct ThreadSpecificData {
-
   int initialized;
 
+  Tcl_Interp *interp;
+
   /*
    * Interpose on default external entity loader
    */
 
+  TclXMLlibxml2Info *current;
   xmlExternalEntityLoader defaultLoader;
 
 } ThreadSpecificData;
@@ -127,9 +130,6 @@ Tclxml_libxml2_Init (interp)
   }
 #endif
 
-  tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
-  tsdPtr->initialized = 1;
-
   classinfo = (TclXML_ParserClassInfo *) ckalloc(sizeof(TclXML_ParserClassInfo));
   classinfo->name = Tcl_NewStringObj("libxml2", -1);
   classinfo->create = ReaderCreate;
@@ -160,11 +160,25 @@ Tclxml_libxml2_Init (interp)
   xmlSubstituteEntitiesDefault(1);
 
   /*
-   * 6 will load external entities.
-   * 0 will not.
    * TODO: provide configuration option for setting this value.
    */
-  xmlLoadExtDtdDefaultValue = 6;
+  xmlLoadExtDtdDefaultValue |= 1;
+  xmlLoadExtDtdDefaultValue |= XML_COMPLETE_ATTRS;
+
+  /*
+   * Override default entity loader so we can implement callbacks
+   */
+
+  tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
+
+  if (!tsdPtr->initialized) {
+    tsdPtr->initialized = 1;
+    tsdPtr->interp = interp;
+
+    tsdPtr->current = NULL;
+    tsdPtr->defaultLoader = xmlGetExternalEntityLoader();
+    xmlSetExternalEntityLoader(TclXMLlibxml2ExternalEntityLoader);
+  } /* only need to init the library once per process */
 
   /* Setting the variable is insufficient - must create namespace too. */
   if (Tcl_VarEval(interp, "namespace eval ::xml::libxml2 {variable libxml2version ", xmlParserVersion, "}\n", NULL) != TCL_OK) {
@@ -173,11 +187,6 @@ Tclxml_libxml2_Init (interp)
 
   Tcl_MutexUnlock(&libxml2);
 
-  /*
-   tsdPtr->contexts = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
-  Tcl_InitHashTable(tsdPtr->contexts, TCL_ONE_WORD_KEYS);
-   */
-
   TclXML_libxml2_InitDocObj(interp);
 
   /*
@@ -200,6 +209,18 @@ Tclxml_libxml2_Init (interp)
 
   return TCL_OK;
 }
+
+/*
+ * TclXML/libxml2 is made safe by preventing the use of the default
+ * external entity loader.
+ */
+
+int
+Tclxml_libxml2_SafeInit (interp)
+     Tcl_Interp *interp;	/* Interpreter to initialise */
+{
+  return Tclxml_libxml2_Init(interp);
+}
 
 /*
  *----------------------------------------------------------------------------
@@ -250,22 +271,22 @@ ReaderCreate(interp, xmlinfo)
   info->reader = xmlNewTextReader(inputPtr, NULL);
   if (info->reader == NULL) {
 	Tcl_MutexUnlock(&libxml2);
-	
+
 	Tcl_Free((char *) info);
 	Tcl_SetResult(interp, "unable to create XML reader", NULL);
 	return NULL;
   }
    xmlTextReaderSetStructuredErrorHandler(info->reader,
-										  (xmlStructuredErrorFunc) TclXML_libxml2_ErrorHandler,
-										  NULL);
+					  (xmlStructuredErrorFunc) TclXML_libxml2_ErrorHandler,
+					  NULL);
 
   Tcl_MutexUnlock(&libxml2);
-  
+
   info->docObjPtr = NULL;
   info->keep = TCLXML_LIBXML2_DOCUMENT_IMPLICIT;
   info->scope = (Tcl_HashTable *) Tcl_Alloc(sizeof(Tcl_HashTable));
   Tcl_InitHashTable(info->scope, TCL_STRING_KEYS);
-  
+
   return (ClientData) info;
 }
 
@@ -292,16 +313,16 @@ TclXMLlibxml2Delete(clientData)
   TclXMLlibxml2Info *info = (TclXMLlibxml2Info *) clientData;
 
   if (info->reader) {
-	xmlFreeTextReader(info->reader);
+    xmlFreeTextReader(info->reader);
   }
   if (info->docObjPtr) {
     Tcl_DecrRefCount(info->docObjPtr);
   }
   if (info->preserve) {
-	Tcl_DecrRefCount(info->preserve);
+    Tcl_DecrRefCount(info->preserve);
   }
   if (info->preservens) {
-	Tcl_DecrRefCount(info->preservens);
+    Tcl_DecrRefCount(info->preservens);
   }
   Tcl_DeleteHashTable(info->scope);
   Tcl_Free((char *) info->scope);
@@ -333,10 +354,10 @@ ClientData clientData;
   TclXML_Info *xmlinfo = (TclXML_Info *) clientData;
 
   if (xmlinfo->clientData == NULL) {
-	xmlinfo->clientData = (ClientData) ReaderCreate(xmlinfo->interp, xmlinfo);
-	if (xmlinfo->clientData == NULL) {
-	  return TCL_ERROR;
-	}
+    xmlinfo->clientData = (ClientData) ReaderCreate(xmlinfo->interp, xmlinfo);
+    if (xmlinfo->clientData == NULL) {
+      return TCL_ERROR;
+    }
   }
 
   return TCL_OK;
@@ -367,13 +388,15 @@ ReaderParse(clientData, data, len, final)
      int len;
      int final;
 {
+  ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
   TclXMLlibxml2Info *info = (TclXMLlibxml2Info *) clientData;
-  Tcl_Obj *nameObj, *nsObj, *valueObj, *attrsObj, *errObj;
-  char *baseuri, *encoding, *name, *ns, *value;
+  Tcl_Obj *nameObj, *nsObj, *nsdeclObj, *valueObj, *attrsObj, *errObj, *baseuriObj, *sysidObj, *extidObj;
+  const char *baseuri, *encoding, *name, *ns, *value;
   xmlChar **preservens = NULL;
-  int ret, i, listlen;
+  int ret, result, i, listlen, options = 0, empty;
 
   xmlNodePtr nodePtr;
+  xmlEntityPtr entityPtr = NULL;
 
   if (final == 0) {
     Tcl_SetResult(info->interp, "partial input not yet supported", NULL);
@@ -415,13 +438,27 @@ ReaderParse(clientData, data, len, final)
 
   TclXML_libxml2_ResetError(info->interp);
 
+  options |= XML_PARSE_NOCDATA;
+
+  tsdPtr->current = info;
+
   Tcl_MutexLock(&libxml2);
 
+  if (info->xmlinfo->expandinternalentities) {
+    options |= XML_PARSE_NOENT;
+    xmlSubstituteEntitiesDefault(1);
+  } else {
+    xmlSubstituteEntitiesDefault(0);
+  }
+  if (info->xmlinfo->nowhitespace) {
+    options |= XML_PARSE_NOBLANKS;
+  }
+
   if (xmlReaderNewMemory(info->reader,
 			 data, len,
 			 baseuri,
 			 encoding,
-			 XML_PARSE_NOCDATA) != 0) {
+			 options) != 0) {
     Tcl_MutexUnlock(&libxml2);
     if (preservens) {
       int i;
@@ -430,14 +467,17 @@ ReaderParse(clientData, data, len, final)
       }
       Tcl_Free((char *) preservens);
     }
+    tsdPtr->current = NULL;
     Tcl_SetResult(info->interp, "unable to prepare parser", NULL);
     return TCL_ERROR;
   }
 
   if (info->preserve) {
-    if (xmlTextReaderPreservePattern(info->reader,
-				     Tcl_GetStringFromObj(info->preserve, NULL),
-				     (const xmlChar **) preservens) < 0) {
+    int preserveret;
+    preserveret = xmlTextReaderPreservePattern(info->reader,
+					       Tcl_GetStringFromObj(info->preserve, NULL),
+					       (const xmlChar **) preservens);
+    if (preserveret < 0) {
       Tcl_MutexUnlock(&libxml2);
       if (preservens) {
 	int i;
@@ -446,6 +486,13 @@ ReaderParse(clientData, data, len, final)
 	}
 	Tcl_Free((char *) preservens);
       }
+      tsdPtr->current = NULL;
+      Tcl_ResetResult(info->interp);
+      Tcl_AppendResult(info->interp,
+		       "preparation for parsing failed: unable to preserve pattern \"",
+		       Tcl_GetStringFromObj(info->preserve, NULL),
+		       "\"",
+		       NULL);
       return TCL_ERROR;
     }
   }
@@ -453,10 +500,11 @@ ReaderParse(clientData, data, len, final)
   for (ret = xmlTextReaderRead(info->reader);
        ret == 1;
        ret = xmlTextReaderRead(info->reader)) {
+    result = TCL_OK;
     switch (xmlTextReaderNodeType(info->reader)) {
     case XML_READER_TYPE_ELEMENT:
-      name = xmlTextReaderLocalName(info->reader);
-      ns = xmlTextReaderNamespaceUri(info->reader);
+      name = xmlTextReaderConstLocalName(info->reader);
+      ns = xmlTextReaderConstNamespaceUri(info->reader);
       Tcl_MutexUnlock(&libxml2);
       if (name != NULL) {
 	nameObj = Tcl_NewStringObj(name, -1);
@@ -470,6 +518,7 @@ ReaderParse(clientData, data, len, final)
 	nsObj = Tcl_NewObj();
       }
       Tcl_IncrRefCount(nsObj);
+
       attrsObj = Tcl_NewObj();
       Tcl_IncrRefCount(attrsObj);
       Tcl_MutexLock(&libxml2);
@@ -478,24 +527,36 @@ ReaderParse(clientData, data, len, final)
 	  Tcl_Obj *itemObj;
 
 	  itemObj = Tcl_NewObj();
-	  Tcl_SetStringObj(itemObj, xmlTextReaderLocalName(info->reader), -1);
+	  Tcl_SetStringObj(itemObj, xmlTextReaderConstLocalName(info->reader), -1);
 	  Tcl_ListObjAppendElement(info->interp, attrsObj, itemObj);
-	  itemObj = Tcl_NewStringObj(xmlTextReaderValue(info->reader), -1);
+	  itemObj = Tcl_NewStringObj(xmlTextReaderConstValue(info->reader), -1);
 	  Tcl_ListObjAppendElement(info->interp, attrsObj, itemObj);
+
 	  while (xmlTextReaderMoveToNextAttribute(info->reader) == 1) {
-	    itemObj = Tcl_NewStringObj(xmlTextReaderLocalName(info->reader), -1);
+	    itemObj = Tcl_NewStringObj(xmlTextReaderConstLocalName(info->reader), -1);
 	    Tcl_ListObjAppendElement(info->interp, attrsObj, itemObj);
-	    itemObj = Tcl_NewStringObj(xmlTextReaderValue(info->reader), -1);
+	    itemObj = Tcl_NewStringObj(xmlTextReaderConstValue(info->reader), -1);
 	    Tcl_ListObjAppendElement(info->interp, attrsObj, itemObj);
 	  }
 	}
       }
+      empty = xmlTextReaderIsEmptyElement(info->reader);
       Tcl_MutexUnlock(&libxml2);
 
+      nsdeclObj = Tcl_NewObj();
+      Tcl_IncrRefCount(nsdeclObj);
+
       TclXML_ElementStartHandler(info->xmlinfo,
 				 nameObj,
 				 nsObj,
-				 attrsObj, Tcl_NewObj());
+				 attrsObj, nsdeclObj);
+
+      Tcl_DecrRefCount(nsdeclObj);
+
+      if (empty) {
+	TclXML_ElementEndHandler(info->xmlinfo,
+				 nameObj);
+      }
 
       Tcl_DecrRefCount(nameObj);
       Tcl_DecrRefCount(nsObj);
@@ -503,8 +564,8 @@ ReaderParse(clientData, data, len, final)
       break;
 
     case XML_READER_TYPE_END_ELEMENT:
-      name = xmlTextReaderLocalName(info->reader);
-      ns = xmlTextReaderNamespaceUri(info->reader);
+      name = xmlTextReaderConstLocalName(info->reader);
+      ns = xmlTextReaderConstNamespaceUri(info->reader);
       Tcl_MutexUnlock(&libxml2);
 
       if (name != NULL) {
@@ -531,7 +592,7 @@ ReaderParse(clientData, data, len, final)
     case XML_READER_TYPE_CDATA:
     case XML_READER_TYPE_WHITESPACE:
     case XML_READER_TYPE_SIGNIFICANT_WHITESPACE:
-      value = xmlTextReaderValue(info->reader);
+      value = xmlTextReaderConstValue(info->reader);
       Tcl_MutexUnlock(&libxml2);
       if (value != NULL) {
 	valueObj = Tcl_NewStringObj(value, -1);
@@ -547,20 +608,21 @@ ReaderParse(clientData, data, len, final)
       break;
 
     case XML_READER_TYPE_COMMENT:
-      value = xmlTextReaderValue(info->reader);
+      value = xmlTextReaderConstValue(info->reader);
       Tcl_MutexUnlock(&libxml2);
       if (value != NULL) {
 	valueObj = Tcl_NewStringObj(value, -1);
       } else {
 	valueObj = Tcl_NewObj();
       }
+
       TclXML_CommentHandler(info->xmlinfo,
 			    valueObj);
       break;
 
     case XML_READER_TYPE_PROCESSING_INSTRUCTION:
-      name = xmlTextReaderName(info->reader);
-      value = xmlTextReaderValue(info->reader);
+      name = xmlTextReaderConstName(info->reader);
+      value = xmlTextReaderConstValue(info->reader);
       Tcl_MutexUnlock(&libxml2);
       if (name != NULL) {
 	nameObj = Tcl_NewStringObj(name, -1);
@@ -572,28 +634,55 @@ ReaderParse(clientData, data, len, final)
       } else {
 	valueObj = Tcl_NewObj();
       }
+
       TclXML_ProcessingInstructionHandler(info->xmlinfo,
 					  nameObj,
 					  valueObj);
       break;
 
     case XML_READER_TYPE_ENTITY_REFERENCE:
-      name = xmlTextReaderName(info->reader);
+      name = xmlTextReaderConstName(info->reader);
+      baseuri = xmlTextReaderConstBaseUri(info->reader);
+      entityPtr = xmlGetDocEntity(xmlTextReaderCurrentDoc(info->reader),
+				  name);
       Tcl_MutexUnlock(&libxml2);
-      if (name != NULL) {
-	nameObj = Tcl_NewStringObj(value, -1);
-      } else {
-	nameObj = Tcl_NewObj();
+
+      nameObj = Tcl_NewStringObj(name, -1);
+      Tcl_IncrRefCount(nameObj);
+      baseuriObj = Tcl_NewStringObj(baseuri, -1);
+      Tcl_IncrRefCount(baseuriObj);
+      sysidObj = Tcl_NewStringObj(entityPtr->SystemID, -1);
+      Tcl_IncrRefCount(sysidObj);
+      extidObj = Tcl_NewStringObj(entityPtr->ExternalID, -1);
+      Tcl_IncrRefCount(extidObj);
+
+      result = TclXML_ExternalEntityRefHandler(info->xmlinfo,
+					       nameObj,
+					       baseuriObj,
+					       sysidObj,
+					       extidObj);
+
+      Tcl_MutexLock(&libxml2);
+
+      Tcl_DecrRefCount(nameObj);
+      Tcl_DecrRefCount(baseuriObj);
+      Tcl_DecrRefCount(sysidObj);
+      Tcl_DecrRefCount(extidObj);
+
+      if (result == TCL_ERROR || result == TCL_BREAK) {
+	Tcl_MutexUnlock(&libxml2);
+	xmlTextReaderClose(info->reader);
+	break;
       }
-      TclXML_ExternalEntityRefHandler(info->xmlinfo,
-				      nameObj,
-				      Tcl_NewObj(),
-				      Tcl_NewObj(),
-				      Tcl_NewObj());
+      Tcl_MutexUnlock(&libxml2);
+
+      break;
 
     case XML_READER_TYPE_DOCUMENT_TYPE:
-      name = xmlTextReaderName(info->reader);
-      nodePtr = xmlTextReaderCurrentNode(info->reader);
+      /* these are not used... at present
+	name = xmlTextReaderName(info->reader);
+	nodePtr = xmlTextReaderCurrentNode(info->reader);
+      */
       Tcl_MutexUnlock(&libxml2);
 
     default:
@@ -612,22 +701,25 @@ ReaderParse(clientData, data, len, final)
     Tcl_Free((char *) preservens);
   }
 
-  if (ret != 0) {
+  if (ret != 0 || result != TCL_OK) {
     errObj = TclXML_libxml2_GetErrorObj(info->interp);
     if (errObj) {
       Tcl_SetObjResult(info->interp, errObj);
     } else {
       Tcl_SetResult(info->interp, "parsing error", NULL);
     }
+    tsdPtr->current = NULL;
+
     return TCL_ERROR;
   }
 
   info->docObjPtr = TclXML_libxml2_CreateObjFromDoc(xmlTextReaderCurrentDoc(info->reader));
 
   TclXML_libxml2_DocKeep(info->docObjPtr, info->keep);
-  
+
   /* TODO: errObjPtr may contain warnings, flush them through */
   
+  tsdPtr->current = NULL;
   return TCL_OK;
 }
 
@@ -658,14 +750,14 @@ TclXMLlibxml2Configure(clientData, optionPtr, valuePtr)
   char *value;
   CONST84 char *Options[] = {
     "-keep",
-	"-retainpath",
-	"-retainpathns",
+    "-retainpath",
+    "-retainpathns",
     NULL
   };
   enum Options {
     OPTION_KEEP,
-	OPTION_RETAINPATH,
-	OPTION_RETAINPATHNS
+    OPTION_RETAINPATH,
+    OPTION_RETAINPATHNS
   };
   CONST84 char *KeepOptions[] = {
     "normal",
@@ -822,6 +914,26 @@ TclXMLlibxml2Get(clientData, objc, objv)
  *----------------------------------------------------------------------------
  */
 
+static xmlParserInputPtr
+Result2ParserInput(interp, ctxt, URL)
+     Tcl_Interp *interp;
+     xmlParserCtxtPtr ctxt;
+     const char *URL;
+{
+  xmlParserInputPtr inputPtr = NULL;
+
+  /* build our own xmlParserInput from returned data */
+  inputPtr = xmlNewStringInputStream(ctxt, Tcl_GetStringFromObj(Tcl_GetObjResult(interp), NULL));
+  if (inputPtr == NULL) {
+    Tcl_SetResult(interp, "unable to create input stream", NULL);
+    Tcl_BackgroundError(interp);
+    return NULL;
+  }
+  inputPtr->filename = (char *) xmlCanonicPath((const xmlChar *) URL);
+
+  return inputPtr;
+}
+
 static xmlParserInputPtr
 TclXMLlibxml2ExternalEntityLoader(URL, ID, ctxt)
     const char *URL;
@@ -829,39 +941,44 @@ TclXMLlibxml2ExternalEntityLoader(URL, ID, ctxt)
     xmlParserCtxtPtr ctxt;
 {
   ThreadSpecificData *tsdPtr = (ThreadSpecificData *) Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
-  TclXMLlibxml2Info *info = NULL;
-#if 0
-  Tcl_HashEntry *entryPtr;
+  TclXMLlibxml2Info *info;
+  Tcl_Interp *interp;
+  int result;
 
-  entryPtr = Tcl_FindHashEntry(tsdPtr->contexts, (ClientData) ctxt);
-  if (!entryPtr) {
-    /* Background error */
-    return NULL;
+  info = tsdPtr->current;
+
+  if (info) {
+    result = TclXML_ExternalEntityRefHandler(info->xmlinfo, NULL, NULL,
+					     Tcl_NewStringObj(URL, -1),
+					     Tcl_NewStringObj(ID, -1));
+    interp = info->interp;
+  } else {
+    result = TclXML_ExternalEntityRefHandler(NULL, NULL, NULL,
+					     Tcl_NewStringObj(URL, -1),
+					     Tcl_NewStringObj(ID, -1));
+    interp = tsdPtr->interp;
   }
-  info = (TclXMLlibxml2Info *) Tcl_GetHashValue(entryPtr);
-#endif
-  
-  if (!info) {
+
+  switch (result) {
+  case TCL_OK:
+    return Result2ParserInput(interp, ctxt, URL);
+
+  case TCL_BREAK:
+    return NULL;
+
+  case TCL_CONTINUE:
+    break;
+
+  case TCL_ERROR:
+  case TCL_RETURN:
+  default:
+    Tcl_BackgroundError(interp);
     return NULL;
   }
 
-  if (info->xmlinfo->entity) {
-    (info->xmlinfo->entity)(info->interp, info->xmlinfo->clientData, NULL, NULL, Tcl_NewStringObj(URL, -1), Tcl_NewStringObj(ID, -1));
-  } else if (info->xmlinfo->entitycommand) {
-    Tcl_Obj *cmdPtr = Tcl_DuplicateObj(info->xmlinfo->entitycommand);
-    
-    if (URL) {
-      Tcl_ListObjAppendElement(info->interp, cmdPtr, Tcl_NewStringObj(URL, -1));
-    } else {
-      Tcl_ListObjAppendElement(info->interp, cmdPtr, Tcl_NewListObj(0, NULL));
-    }
-    if (ID) {
-      Tcl_ListObjAppendElement(info->interp, cmdPtr, Tcl_NewStringObj(ID, -1));
-    } else {
-      Tcl_ListObjAppendElement(info->interp, cmdPtr, Tcl_NewListObj(0, NULL));
-    }
+  if (Tcl_IsSafe(interp)) {
+    return NULL;
   } else {
     return (tsdPtr->defaultLoader)(URL, ID, ctxt);
   }
-  return NULL; /* never reached */
 }
